# Natural Language Toolkit: Unit Tests
#
# Copyright (C) 2001-2009 NLTK Project
# Author: Edward Loper <edloper@gradient.cis.upenn.edu>
# URL: <http://www.nltk.org/>
# For license information, see LICENSE.TXT
#
# $Id: __init__.py 7460 2009-01-29 01:06:02Z StevenBird1 $

"""
Unit tests for the NLTK modules.  These tests are intented to ensure
that changes that we make to NLTK's code don't accidentally introduce
bugs.

Use doctest_driver.py to run the tests::

  doctest_driver.py --help

NB. Popular options for NLTK documentation are::

  --ellipsis --normalize_whitespace

"""
